package com.sodius.mdw.log;

import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import com.sodius.mdw.core.util.log.AbstractLogger;
import com.sodius.mdw.core.util.log.Level;
import com.sodius.mdw.log.reporter.LogReporter;

/**
 * @author Mickael Albert
 * Record messages and EventStatus in memory
 * Send recorded Status to reporters
 */
public class EventStackLogger extends AbstractLogger {

	/**
	 * Default EventCode when logging event which is not an EventStatus
	 *
	 */
	public static final class UnknownCode implements EventCode {
		
		private final String description;
		
		private UnknownCode(String description) {
			this.description = description;
		}

		public int getCode() {
			return 0;
		}

		public String getName() {
			return "UNKNOWN";
		}

		public String getDescription() {
			return description;
		}

		public EventKind getKind() {
			return EventKind.MISC;
		}
	};
	
	protected final List<EventStatus> stack = new LinkedList<EventStatus>();
	private final List<LogReporter> reporters = new ArrayList<LogReporter>(); 
	
	@Override
	protected void log(Level level, Object message, Throwable error) {
		if(!isLoggable(level))
			return;
		if(message instanceof EventStatus)
		{
			log((EventStatus)message);
		}
		else
		{
			EventSeverity severity=null;
			if(Level.DEBUG==level)
				severity=EventSeverity.DEBUG;
			else if(Level.ERROR==level)
				severity=EventSeverity.ERROR;
			else if(Level.INFO==level)
				severity=EventSeverity.INFO;
			else if(Level.WARNING==level)
				severity=EventSeverity.WARNING;
			log(EventStatus.createStatus(new UnknownCode(message.toString()), severity, error, null));
		}
	}
	
	private void log(EventStatus status)
	{
		stack.add(status);
	}

	/**
	 * Add a new reporter in the list of registered reporter.
	 * @param reporter
	 * @see EventStackLogger#flush()
	 */
	public void registerReporter(LogReporter reporter)
	{
		reporters.add(reporter);
	}
	
	/**
	 * Call each reporter to report the stack of recorded Status
	 * Finally clear this stack of status 
	 * onlyWithStack=false
	 * @throws IOException
	 */
	public void flush() throws IOException
	{
		flush(false);
	}
	
	/**
	 * Call each reporter to report the stack of recorded Status
	 * Finally clear this stack of status 
	 * @param onlyWithStack if true, call reporters only if stack non empty
	 * @throws IOException
	 */
	public void flush(boolean onlyWithStack) throws IOException
	{
		if(onlyWithStack && this.stack.isEmpty())
			return;
		for (LogReporter logReporter : reporters) {
			logReporter.report(this.stack);
		}
		this.stack.clear();
	}
	
	@Override
	protected void finalize() throws Throwable {
		if(!(reporters.isEmpty() || stack.isEmpty()))
			flush();
		super.finalize();
	}
	
}
