package com.sodius.mdw.log.reporter;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.Writer;
import java.util.List;

import com.sodius.mdw.log.EventSeverity;
import com.sodius.mdw.log.EventStatus;
import com.sodius.mdw.log.reporter.tree.SubTreeGenerator;
import com.sodius.mdw.log.reporter.tree.Tree;


/**
 * @author Mickael Albert
 * 
 * Generate a tree report in HTML
 * Each SubTreeGenerator will generate a tree in the HTML
 * 
 * HTML Tree description:
 * <OL class="tree"> //Root element defining a tree
 * <LI>  //Tree
 * 	<LABEL class='folder' ../> //Tree name, for describing an interactive folder
 * 		<INPUT type="checkbox" ... id="treeID{n}" /> // ie "treeID0" with n the nth element. Used as workaround for short and compatible css
 * 
 * 		<OL>
 * 			... containment, as other trees
 * 		</OL>
 *
 */
public class HTMLTreeLogReporter  extends TextLogReporter{

	private final SubTreeGenerator[] treeGens;
	private final File ressourceFolder;
	private final String title;
	protected static final String HTML_TREE = "OL";
	protected static final String HTML_SUBTREE = "LI";
	protected static final String HTML_LABEL = "LABEL";
	protected static final String HTML_SELECTOR = "INPUT";
	protected static final String HTML_CLASS_FILE = "file";
	protected static final String HTML_CLASS_ROOT_TREE = "tree";
	protected int treeID=0;
	
	/**
	 * @param minimumSeverity
	 * @param maximumSeverity
	 * @param writer
	 * @param header
	 * @param footer
	 * @param title
	 * @param ressourceFolder Folder containing all css required for this HTML. The HTML will directly reference those files.
	 * @param treeGens list of SubTreeGenerator for generating html trees
	 * @throws IOException
	 */
	public HTMLTreeLogReporter(EventSeverity minimumSeverity,
			EventSeverity maximumSeverity, Writer writer, String header, String footer, String title, File ressourceFolder, SubTreeGenerator... treeGens) throws IOException {
		super(minimumSeverity, maximumSeverity, writer, header, footer);
		this.treeGens=treeGens;
		this.title=title;
		this.ressourceFolder=ressourceFolder;
	}

	@Override
	public void report(List<EventStatus> statusList) throws IOException {
		Tree[] rootTrees = new Tree[treeGens.length];
		for (int i = 0; i < treeGens.length; i++) {
			rootTrees[i] = Tree.getTree(statusList, treeGens[i], minimumSeverity, maximumSeverity);
		}
		writeHeaderHTML();
		for (int i = 0; i < rootTrees.length; i++) {
			writeSubTitle("Event tree based on "+treeGens[i].getSortName()); //TODO: encore sortName in HTML format
			writeSubTree(rootTrees[i], "", true);
		}
		writeFooterHTML();
		writer.flush();
	}
	
	protected void writeSubTitle(String subTitle) throws IOException
	{
		writer.append("<T2>"+subTitle+"</T2>"); 
	}
	
	@Override
	protected void writerHeader() throws IOException {
		//Nothing
	}
	
	@Override
	protected void writerFooter() throws IOException {
		//Nothing
	}
	
	protected void writeHeaderHTML() throws IOException
	{
		writer.append("<HTML>\n");
		writer.append("<HEAD>\n");
		
		//Search all CSS
		if(ressourceFolder!=null && ressourceFolder.isDirectory())
		{
			FilenameFilter cssFilter = new FilenameFilter() {
				public boolean accept(File dir, String name) {
					return name.endsWith(".css");
				}
			};
			for (File file : ressourceFolder.listFiles(cssFilter)) {
				writer.append("<!--[if gte IE 9 ]><link rel=\"stylesheet\" type=\"text/css\" href=\"" + file.toURL().toString() + "\" media=\"screen\"><![endif]-->\n");
				writer.append("<!--[if !IE]>--><link rel=\"stylesheet\" type=\"text/css\" href=\"" + file.toURL().toString() + "\" media=\"screen\"><!--<![endif]-->\n");
			}
		}
		if(title!=null)
		{
			writer.append("	<TITLE>" + title +"</TITLE>\n");
			
		}
		writer.append(	"</HEAD>\n" +
						"<BODY>\n");
		writer.append(header);
	}
	
	protected void writeFooterHTML() throws IOException
	{
		writer.append(footer);
		writer.append(	"</BODY>\n" +
						"</HTML>\n");
	}
	
	protected void writeTree(Tree tree, String tab) throws IOException
	{
		//LABEL
		writer.append("<" + HTML_LABEL + " class='folder'" + (tree.getComment()!=null?" title=\""+tree.getComment() + "\"":"") 
				+ " for=\"treeID" + treeID + "\">[" + tree.getAllContainedStatusNumber() + "]" +  tree.getContent() + "</" + HTML_LABEL + ">");
		//BEHAVIOR
		writer.append("<" + HTML_SELECTOR + " type=\"checkbox\" style=\"display:none\" id=\"treeID" + treeID++ + "\"/>\n");
		writeSubTree(tree, tab, false);
	}
	
	protected void writeSubTree(Tree tree, String tab, boolean isRoot) throws IOException
	{
		//TREE
		writer.append(tab+"<" + HTML_TREE + (isRoot?" class=\""+HTML_CLASS_ROOT_TREE+"\"":"") + ">\n");
		//subtrees
		String innerTab=tab+'\t';
		for (Tree subtree : tree.getChildren()) {
			writer.append(innerTab + "<"+HTML_SUBTREE+">");
			writeTree(subtree, innerTab);
			writer.append(innerTab+"</" + HTML_SUBTREE + ">\n");
		}
		//Status
		for (EventStatus status : tree.getContainedStatus()) {
			writer.append(innerTab+"<"+HTML_SUBTREE+">");
			writeStatus(status, innerTab);
			writer.append(innerTab+"</"+HTML_SUBTREE+">\n");
		}
		writer.append(tab + "</" + HTML_TREE + ">\n");
	}
	
	protected void writeStatus(EventStatus status, String tab) throws IOException
	{
		writeStatusDescription(status, tab);
		writer.append('\n');
		String innerTab=tab+"\t";
		if(status.getException()!=null)
		{
			writer.append(tab+"<BR>Caused by:<BR>"+innerTab);
			writeException(status.getException(), innerTab);
		}
		if(status.getChildren()!=null && status.getChildren().length>0)
		{
			writer.append(tab+"<BR>Childrens:");
			for (int i = 0; i < status.getChildren().length; i++) {
				writer.append("<BR>"+innerTab);
				if(status.getChildren()[i] instanceof EventStatus)
					writeStatus((EventStatus)status.getChildren()[i], innerTab);
				else
					writer.append(status.getChildren()[i].toString());
			}
		}
	}
	
	protected void writeException(Throwable exception, String tab) throws IOException
	{
		writer.append(exception.toString());
	}
	
	protected void writeStatusDescription(EventStatus status, String tab) throws IOException
	{
		//label title
		writer.append("<LABEL title='"+ status.getEventCode().getKind().getName() + " Event" +
					(status.getContext()!=null?" at "+status.getContext().toString():"") +
					"'>");
		// description
		writer.append("<b>{"+status.getEventSeverity()+"}</b> "+ status.getMessage()+
				(status.getContext()!=null?" at <i>"+status.getContext().getSimpleName()+"</i>":"")+
				"</LABEL>");
	}
	
}
