package com.sodius.mdw.log.reporter.tree;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import com.sodius.mdw.log.EventStatus;

/**
 * @author Mickael Albert
 *
 * Generates subtrees 
 */
public abstract class SubTreeGenerator
	{
		public static final String EMPTY_GROUP_NAME="MISC";
		protected final boolean displayUnnamedGroup;
		
		/**
		 * set displayUnnamedGroup = true
		 */
		public SubTreeGenerator() {
			this.displayUnnamedGroup = true;
		}
		
		/**
		 * @param displayUnnamedGroup if true, create a Tree for containing unnamed trees
		 */
		public SubTreeGenerator(boolean displayUnnamedGroup) {
			this.displayUnnamedGroup = displayUnnamedGroup;
		}

		/**
		 * @param tree
		 * @return a list of SubTrees. Empty list if not possible to split tree
		 */
		public Collection<Tree> generateSubTrees(Tree tree)
		{
			if(!canGenerateSubTrees(tree))
				return Collections.EMPTY_LIST;
			Map<String, Tree> groupID2Tree = new HashMap<String, Tree>();
			int groupLevel=tree.getGroupLevel()+1;
			for (EventStatus child : tree.getContainedStatus()) {
				String groupID = getGroupID(child, groupLevel);
				if(groupID==null)
				{
					if(!displayUnnamedGroup)
						continue;
					groupID=EMPTY_GROUP_NAME;
				}
				if(!groupID2Tree.containsKey(groupID))
					groupID2Tree.put(groupID, createTree(groupID,groupLevel, child));
				
				groupID2Tree.get(groupID).getContainedStatus().add(child);
			}
			tree.reinitContainedStatus();
			tree.getChildren().addAll(groupID2Tree.values());
			return tree.getChildren();
		}
		
		protected Tree createTree(String groupID, int groupLevel, EventStatus status)
		{
			return new Tree(getContent(groupID, groupLevel, status)
							,getComment(groupID, groupLevel, status)
							,getQualifiedName(groupID, groupLevel, status)
							,groupLevel);
		}
		
		/**
		 * @param tree
		 * @return true if the tree can be split in one or more trees with GroupIDs
		 */
		public abstract boolean canGenerateSubTrees(Tree tree);
		
		/**
		 * @param groupID
		 * @param groupLevel
		 * @param status
		 * @return the computed content of a tree representing groupID, groupLevel, and a sampled status
		 */
		protected String getContent(String groupID, int groupLevel, EventStatus status)
		{
			return groupID;
		}
		
		
		/**
		 * @param groupID
		 * @param groupLevel
		 * @param status
		 * @return the computed comment of a tree representing groupID, groupLevel, and a sampled status
		 */
		protected abstract String getComment(String groupID, int groupLevel, EventStatus status);
		
		
		/**
		 * @param groupID
		 * @param groupLevel
		 * @param status
		 * @return the computed QualifiedName of a tree representing groupID, groupLevel, and a sampled status.
		 */
		protected String getQualifiedName(String groupID, int groupLevel, EventStatus status)
		{
			return groupID;
		}
		
		/**
		 * @param status
		 * @param groupLevel
		 * @return the groupID representing this status at the specified groupLevel. Null if not managed by this generator
		 */
		protected abstract String getGroupID(EventStatus status, int groupLevel);
		
		/**
		 * @return the name of the sort used by the SubTreeGenerator
		 */
		public abstract String getSortName();
		
	}
	