package com.sodius.mdw.log.reporter.tree;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import com.sodius.mdw.log.EventSeverity;
import com.sodius.mdw.log.EventStatus;

/**
 * @author Mickael Albert
 *
 * Describes a tree of EventStatus
 * contains subtrees
 * contains status directly attached to this branch
 */
public class Tree
	{
		private final List<Tree> children=new ArrayList<Tree>();
		List<EventStatus> containedStatus=new ArrayList<EventStatus>();
		private final String content;
		private final String comment;
		private final String qualifiedName;
		private final int groupLevel;
		
		/**
		 * @param content The description of the tree
		 * @param comment 
		 * @param qualifiedName The ID of the tree
		 * @param groupLevel
		 */
		Tree(String content, String comment, String qualifiedName, int groupLevel) {
			this.content = content;
			this.comment = comment;
			this.qualifiedName=qualifiedName;
			this.groupLevel=groupLevel;
		}

		/**
		 * @return the distance minus 1 to the rootTree:<BR>-1 is rootTree<BR>0..n are subtrees
		 */
		public int getGroupLevel()
		{
			return groupLevel;
		}

		/**
		 * @return The description/name of the Tree
		 */
		public String getContent() {
			return content;
		}

		public String getComment() {
			return comment;
		}

		/**
		 * @return the ID of the Tree. Could be a concatenation of recursive content
		 */
		public String getQualifiedName() {
			return qualifiedName;
		}
		
		public List<Tree> getChildren() {
			return children;
		}

		public List<EventStatus> getContainedStatus() {
			return containedStatus;
		}
		
		/**
		 *  remove all containedStatus
		 */
		public void reinitContainedStatus() {
			containedStatus=new ArrayList<EventStatus>();
		}	
		
		public static Tree getTree(List<EventStatus> statusList, SubTreeGenerator treeGen, EventSeverity minimumSeverity, EventSeverity maximumSeverity)
		{
			Tree rootTree=new Tree("", "", "",-1);
			for (EventStatus eventStatus : statusList) {
				if(eventStatus.getEventSeverity().compareTo(minimumSeverity)>=0 &&
						eventStatus.getEventSeverity().compareTo(maximumSeverity)<=0)
				{
					rootTree.containedStatus.add(eventStatus);
				}
			}
			LinkedList<Tree> treeToSplit = new LinkedList<Tree>();
			treeToSplit.add(rootTree);
			while(!treeToSplit.isEmpty())
			{
				Tree tree = treeToSplit.getFirst();
				treeToSplit.removeFirst();
				if(treeGen.canGenerateSubTrees(tree))
				{
					treeToSplit.addAll(treeGen.generateSubTrees(tree));
				}
			}
			rootTree.reinitContainedStatus(); // RootTree is just a subtree container. If it still contains status, that means it was reject during subtree generation.
			return rootTree;
		}
		
		/**
		 * @return True if it does not contains any subtrees
		 */
		public boolean isLeaf()
		{
			return getChildren().isEmpty();
		}
		
		/**
		 * @return the number of contained Status recursively through its subtrees
		 */
		public int getAllContainedStatusNumber()
		{
			int total = containedStatus.size();
			for (Tree child : children) {
				total+= child.getAllContainedStatusNumber();
			}
			return total;
		}
		
	}