package com.sodius.rhapsody.toolkit.engine.uml21;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.xmi.XMLResource;

import com.sodius.mdw.core.CoreException;
import com.sodius.mdw.core.eval.Evaluable;
import com.sodius.mdw.core.eval.EvaluationContext;
import com.sodius.mdw.core.eval.project.Project;
import com.sodius.mdw.core.model.MDWObject;
import com.sodius.mdw.core.model.Model;
import com.sodius.mdw.core.model.io.ConnectorDescriptor;
import com.sodius.mdw.core.util.progress.ProgressMonitor;
import com.sodius.mdw.metamodel.rhapsody.util.Counter;
import com.sodius.mdw.metamodel.rhapsody.util.IRhapsodyConstants;
import com.sodius.mdw.metamodel.uml21.Namespace;
import com.sodius.mdw.metamodel.uml21.io.ea.Options;
import com.sodius.mdw.metamodel.uml21.io.ea.utils.InternalOptions;
import com.sodius.mdw.metamodel.uml21.resource.XMI2UMLResource;
import com.sodius.rhapsody.toolkit.engine.EngineContext;
import com.sodius.rhapsody.toolkit.properties.RhapsodyXMIToolkitParameters;
import com.sodius.rhapsody.toolkit.properties.UMLVersionKind;
import com.sodius.rhapsody.toolkit.util.thread.RhapsodyAPIExec;
import com.telelogic.rhapsody.core.IRPApplication;

public class UML21ImportRunner extends UML21EngineRunner {
    private final static String DIAGRAM_EXTENSION = "di";
    private IRPApplication rhpApp = null;

    public UML21ImportRunner(EngineContext context, String toolkitDirectory, IRPApplication rhpApp) {
        super(context, toolkitDirectory);
        this.rhpApp = rhpApp;
    }

    @Override
    protected File getTransformationPluginPath() {
        return new File(getMDWorkbenchFolder(), getRhapsodyPlugin(getMDWorkbenchFolder(), "com.sodius.mdw.UML21toRhapsody_"));
    }

    public static ConnectorDescriptor getUMLModelConnector(RhapsodyXMIToolkitParameters parameters) throws CoreException {

    	if (parameters.getFormat() == UMLVersionKind.uml21EA) {
    		ConnectorDescriptor reader = parameters.getSelectedMetamodel().getModelReaderDescriptor(UMLVersionKind.uml21EA.readerName);
    		reader.setProperty(Options.OPTION_XMI_READ_EXTENSIONS_SUBSET, "properties.documentation,documentation.value");
    		reader.setProperty(InternalOptions.OPTION_XMI_READ_POST_FIX_OPERATIONS, "FIX_ILLEGAL_TRANSITIONS");
    	} else {
    		try {
    			ConnectorDescriptor reader = parameters.getSelectedMetamodel().getModelReaderDescriptor("XMI");
    			reader.setProperty("parseEAExtensions", true);
    			reader.setProperty("parseDiagrams", true);
    		} catch (CoreException e) {
    		}
    	}
    	ConnectorDescriptor selectedReader = parameters.getSelectedReader();
	
    	final String forcedReaderName = parameters.getProperty("forcedReader", null);
    	if (forcedReaderName != null) {
    		final ConnectorDescriptor forcedReader = parameters.getSelectedMetamodel().getModelReaderDescriptor(forcedReaderName);
    		if (forcedReader != null) {
    			selectedReader = forcedReader;
    		}
    	}
    	return selectedReader;
    }
    
    public void run(final ProgressMonitor monitor) throws CoreException {
        final Model umlModel;
        if (context.getParameters().getAlreadyReadImportModel() != null) {
            umlModel = context.getParameters().getAlreadyReadImportModel();
        } else {
        	ConnectorDescriptor selectedReader = getUMLModelConnector(context.getParameters());
            umlModel = context.getParameters().getMDWorkbench().getMetamodelManager().getMetamodel("uml21")
                    .readModel(selectedReader != null ? selectedReader.getName() : "XMI", getContext().getParameters().getXMIFilename());
        }
        if (umlModel.getInstances("Diagram").isEmpty()) {
            // check if di file present.
            try {
                File diFile = context.getParameters().getAdditionalDiagramFile(context.getParameters().getXMIFilename());
                if ((diFile != null) && diFile.exists()) {
                    ResourceSet resourceSet = umlModel.getInstances("Element", true).first().eResource().getResourceSet();
                    resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put(DIAGRAM_EXTENSION, XMI2UMLResource.Factory.INSTANCE);
                    Resource rsDi = resourceSet.createResource(URI.createFileURI(diFile.getAbsolutePath()));
                    try {
                        Map<String, Object> options = new HashMap<String, Object>();
                        options.put(XMLResource.OPTION_RECORD_UNKNOWN_FEATURE, Boolean.TRUE);
                        options.put(XMLResource.OPTION_PROCESS_DANGLING_HREF, XMLResource.OPTION_PROCESS_DANGLING_HREF_RECORD);
                        rsDi.load(options);
                    } catch (IOException e) {
                        // TODO: handle exception
                    }
                    Iterator<EObject> content = rsDi.getAllContents();
                    while (content.hasNext()) {
                        EObject object = content.next();
                        if (object instanceof MDWObject) {
                            umlModel.add((MDWObject) object);
                        }
                    }
                }
            } catch (Exception e) {
                context.getOutLogger().info("An exception occured when searching diagrams. Some diagram layouts could be lost.", e);

            }
        }
        final Model rhaModel = context.getParameters().getMDWorkbench().getMetamodelManager().getMetamodel("rhapsody").createModel();
        final String namespaces;
        if (context.getParameters().getAllUnitsAreSelected()) {
            namespaces = "";
        } else {
            namespaces = getUnitsSelection(context.getParameters().getSelectedNamespacesToImport());
        }
        Project project = getProject("UML2rhapsody");
        final String[] language = new String[1];
        RhapsodyAPIExec.synchExec(new Runnable() {
            public void run() {
                language[0] = rhpApp.getLanguage();

            }
        });

        // create properties for rulesets
        final Properties userProperties = context.getParameters().getUserProperties();

        context.getParameters().getMDWorkbench().getConfiguration().setLogger(new EventToPrintLogger(context.getOutLogger(), context.getErrLogger()));
        context.getParameters().getMDWorkbench().getEvaluationManager().evaluate(project, new Evaluable() {

            public void evaluate(EvaluationContext context) throws CoreException {
                List<Object> ruleSetArguments = new ArrayList<Object>();
                // Adding Rhapsody Model
                ruleSetArguments.add(umlModel);
                // Adding UMLModel
                ruleSetArguments.add(rhaModel);
                // Adding additional args
                ruleSetArguments.add(language[0]);

                List<Object> ruleArguments = new ArrayList<Object>();
                // adding namespaces
                ruleArguments.add(namespaces);

                String[] properties = userProperties.keySet().toArray(new String[] {});
                for (String key : properties) {
                    context.setProperty(key, userProperties.getProperty(key));
                }
                context.evaluateRule("com.sodius.mdw.UML21toRhapsody.rulesets.frameworkRuleset", ruleSetArguments, "runPartial", ruleArguments);

            }

        }, monitor);
        Map<String, Object> writingOptions = new HashMap<String, Object>();
        Counter counter = new Counter();
        ConnectorDescriptor rapDesc = context.getParameters().getMDWorkbench().getMetamodelManager().getMetamodel("rhapsody").getModelWriterDescriptor("Rhapsody Application");
        Object lastImportLevel = rapDesc.getProperty("importLevel");
        rapDesc.setProperty("importLevel", "Package");
        writingOptions.put(IRhapsodyConstants.OPTIONS_COUNT_PROCESSED_ELEMENTS, counter);

        final boolean isRhapsodyCL = context.getParameters().isRhapsodyCL();
        if (isRhapsodyCL) {
            ConnectorDescriptor writer = rhaModel.getMetamodel().getModelWriterDescriptor("Rhapsody Application");
            writer.setProperty("importDiagrams", false);
        }

        if (context.getParameters().getIsPlugin()) {
            ConnectorDescriptor writer = rhaModel.getMetamodel().getModelWriterDescriptor("Rhapsody Application");
            writer.setProperty("activeApplication", context.getXMIEngine().getApplication());

        }
        if (context.getParameters().getImportMethod() == RhapsodyXMIToolkitParameters.importMethodKind.ADD) {
            rhaModel.write("Rhapsody Application", "", writingOptions);
            context.getDirectOutLogger().info(counter.size() + " imported elements.\nComplete logs at " + context.getParameters().getLogDirectory());
        } else if (context.getParameters().getImportMethod() == RhapsodyXMIToolkitParameters.importMethodKind.ADD_AND_REPLACE) {
            try {
                monitor.subTask("Merging in the model...");
                RhapsodyWriter.addToModel(context, rhpApp, context.getParameters().getMDWorkbench(), umlModel, rhaModel, namespaces,
                        context.getParameters().getLogDirectory(), writingOptions);
            } catch (Exception e) {
                context.getOutLogger().error("Error when writing into Rhapsody", e);
            } finally {
                context.getDirectOutLogger()
                        .info(counter.size() + " imported elements.\nComplete logs at " + context.getParameters().getLogDirectory());
            }
        } else if (context.getParameters().getImportMethod() == RhapsodyXMIToolkitParameters.importMethodKind.DIFFMERGE) {
            try {
                monitor.subTask("Launching DiffMerge...");
                RhapsodyWriter.launchDiffMerge(context, rhpApp, context.getParameters().getMDWorkbench(), umlModel, rhaModel,
                        context.getParameters().getLogDirectory(), writingOptions);
            } catch (Exception e) {
                context.getOutLogger().error("Error when writing into Rhapsody", e);
            }
        } else {
            throw new CoreException("The method " + context.getParameters().getImportMethod().literal + " is not implemented");
        }
        rapDesc.setProperty("importLevel", lastImportLevel);
    }

    private String getUnitsSelection(List<Object> selectedUnits) {
        String unitsSelection = "";
        for (Object unitObject : selectedUnits) {
            if (unitObject instanceof Namespace) {
                if (unitsSelection.length() > 0) {
                    unitsSelection += ",";
                }
                unitsSelection += ((Namespace) unitObject).getQualifiedName();
            } else if (unitObject instanceof String) {
                if (unitsSelection.length() > 0) {
                    unitsSelection += ",";
                }
                unitsSelection += unitObject;
            }
        }
        return unitsSelection;
    }
}
